﻿using System;
using System.Drawing;   // required for Point


namespace Pololu
{
    /// <summary>
    /// This class provides methods for reading data from the SLO-scope and for accessing
    /// that data.
    /// 
    /// This class stores the SLO-scope readings in an array that is used as a ring buffer.
    /// As a reading comes in, it is placed in the array at index bufferPosition, and
    /// bufferPosition is then incremented.  If bufferPosition equals buffer.Length, it is
    /// reset back to zero (hence the ring buffer).  Index bufferPosition always points to
    /// the oldest data in the buffer (and therefore the location where the next reading
    /// will be stored).  Index (bufferPosition-1) always points to the newest data in
    /// the buffer (note that if bufferPosition is zero, the newest data is at index
    /// (buffer.Length-1), or, equivalently, ((bufferPosition - 1 + buffer.Length) % buffer.Length),
    /// which is the general-case computation that yields the location of the most recent
    /// data.
    /// </summary>
    public class RawData
    {
        /// <summary>
        /// An object representing the slo-scope we are connected to.
        /// </summary>
        private Sloscope sloscope;

        /// <summary>
        /// The readings that we are holding in memory.  Buffer is an array of raw
        /// channel A and channel B readings implemented as a ring buffer.  Timing information
        /// is directly related to the buffer location of the reading, so timing information
        /// does not have to be separately stored.
        /// </summary>
        private Sloscope.Reading[] buffer;

        /// <summary>
        /// This is used as an index that tells us where the next reading is to be inserted
        /// into the buffer.  Data wraps in the ring buffer, so the oldest data is always
        /// at index bufferPosition, and the newest data is always at (bufferPosition-1).
        /// </summary>
        private int bufferPosition = 0;


        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="memorySize">The number of SLO-scope readings that we can store at a single time.</param>
        public RawData(Sloscope scope, int memorySize)
        {
            sloscope = scope;
            buffer = new Sloscope.Reading[memorySize];
            Reset();
        }


        /// <summary>
        /// Property for getting the number of SLO-scope readings that we can store at a single time.
        /// </summary>
        public int Size
        {
            get { return buffer.Length; }
        }


        /// <summary>
        /// Clear the data buffer.
        /// </summary>
        public void Reset()
        {
            for (int i = 0; i < Size; i++)
            {
                buffer[i].channelA = 0;
                buffer[i].channelB = 0;
            }
            bufferPosition = 0;
        }


        /// <summary>
        /// Reads all existing data from the SLO-scope into the data ring buffer and returns
        /// the number readings.
        /// </summary>
        /// <returns>The number of readings.</returns>
        public int ReadExisting()
        {
            int readingsCount = (int)sloscope.read(buffer, Sloscope.ReadMode.Existing,
                (uint)bufferPosition, (uint)(Size - bufferPosition), 1);
            bufferPosition = (bufferPosition + readingsCount) % Size;

            // sloscope.read() isn't set up to write data into a ring buffer, so if we reach the end of the array, we
            // need to perform a second read starting from the beginning of the array to be sure we get all the data
            while (bufferPosition == 0)
            {
                bufferPosition = (int)sloscope.read(buffer, Sloscope.ReadMode.Existing, 0, (uint)buffer.Length, 1);
                readingsCount += bufferPosition;
                bufferPosition = bufferPosition % Size;
            }

            return readingsCount;
        }


        /// <summary>
        /// Converts the specified queue index in to the associated ring buffer index.  An idx value of
        /// 0 corresponds to the oldest data in the buffer and an idx value of (buffer.Length-1)
        /// corresponds to the most recent data in the buffer.
        /// </summary>
        /// <param name="idx">Queue index.</param>
        /// <returns>Ring buffer index.</returns>
        private int GetBufferIdx(int idx)
        {
            int buffIdx = (bufferPosition + idx) % Size;
            if (buffIdx < 0)
                buffIdx += Size;
            return buffIdx;
        }


        /// <summary>
        /// Populates two arrays of points with the requested channel A and channel B data.  If the requested
        /// data is invalid (i.e. would contain the ring-buffer wrap-around point), this method throws an
        /// InvalidOperationException.
        /// </summary>
        /// <param name="startIdx">Index of the data that should go in pointsA[0] and pointsB[0].
        /// This index should be specified as if it were to a queue and not a ring buffer.
        /// A startIdx of 0 corresponds to the oldest data in the buffer and a startIdx of (buffer.Length-1)
        /// corresponds to the most recent data in the buffer.  startIdx can be greater than endIdx.</param>
        /// <param name="interval">Every interval samples from the buffer will be placed into the pointsA
        /// and pointsB arrays.  An interval of 1 requests all samples, an interval of 2 requests every other
        /// sample, etc.  If interval is negative, samples are read backwards in time starting from startIdx.
        /// If interval is positive, samples are read forwards in time from startIdx.</param>
        /// <param name="pointsA">The array that will contain the requested channel A data.  The size of this
        /// array determines the number of points that will be read from the data buffer.</param>
        /// <param name="pointsB">The array that will contain the requested channel B data.  The size of this
        /// array must match the size of pointsA.</param>
        /// <returns>The return value is false if the specified parameters are invalid (i.e.
        /// if the specified parameters result in requesting data earlier than the earliest value
        /// or data later than the latest value).</returns>
        public void GetData(int startIdx, float interval, Point[] pointsA, Point[] pointsB)
        {
            // check if we can return the requested data
            if (pointsA == null || pointsB == null)
                throw new NullReferenceException("RawData.GetData(): Points arrays must not be null.");
            int endIdx = (int)Math.Round(startIdx + (pointsA.Length - 1) * interval);
            if (endIdx >= Size || endIdx < 0)
                throw new InvalidOperationException("RawData.GetData(): Requested data is outside the allowed range.");

            for (int i = 0; i < pointsA.Length; i++)
            {
                Sloscope.Reading reading = buffer[GetBufferIdx(startIdx + (int)Math.Round(interval * i))];
                int x = interval > 0 ? i : pointsA.Length - 1 - i;
                pointsA[i] = new Point(x, reading.channelA);
                pointsB[i] = new Point(x, reading.channelB);
            }
        }


        /// <summary>
        /// Returns the buffer data element requested by the queue index idx.
        /// </summary>
        /// <param name="idx">Queue index of requested buffer data.  An idx value of
        /// 0 corresponds to the oldest data in the buffer and an idx value of (buffer.Length-1)
        /// corresponds to the most recent data in the buffer.</param>
        /// <returns>The requested channel A and channel B data.</returns>
        public Sloscope.Reading GetData(int idx)
        {
            return buffer[GetBufferIdx(idx)];
        }


        /// <summary>
        /// Returns the index of the oldest data from within the specified range of indices 
        /// that satisfies the specified trigger criteria.  The returned index is a queue index
        /// (e.g. 0 corresponds to the oldest data in the buffer).  If the returned index is -1,
        /// the specified data contained no valid trigger indices.  If the specified range of
        /// indices is invalid, this method throws an InvalidOperationException.
        /// </summary>
        /// <param name="endIdx">The queue index of the most recent data to consider as a
        /// trigger point.  An endIdx is (buffer.Length-1) specifies the most recently received data.</param>
        /// <param name="numPoints">The number of points to search.  If a trigger point is found, the
        /// returned index will be in the set [endIdx-numPoints+1, endIdx], but the method will
        /// need to access buffer data at (endIdx-numPoints) in order to determine if 
        /// (endIdx-numPoints+1) is a trigger point.  One consequence of this is that numPoints can
        /// never be equal to buffer.Length.</param>
        /// <param name="triggerLevel">The level of the trigger point in raw data units.  This
        /// value is directly compared to the buffer channel data.</param>
        /// <param name="risingEdge">Trigger off a rising edge if this is true, else trigger off of
        /// a falling edge.</param>
        /// <param name="channelA">Trigger off of channel A data if this is true, else trigger off
        /// of channel B data.</param>
        /// <returns>The queue index of the oldest valid trigger point.  This value is -1 if no trigger point
        /// was found.</returns>
        public int GetTriggerIdx(int endIdx, int numPoints, int triggerLevel, bool risingEdge, bool channelA)
        {
            if (endIdx >= Size || numPoints < 0 || endIdx - numPoints < 0)
                throw new InvalidOperationException("RawData.GetTriggerIdx(): Requested data is outside the allowed range.");

            for (int i = numPoints - 1; i >= 0; i--)
            {
                int newerReading, olderReading;
                int idx = endIdx - i;
                if (channelA)
                {
                    newerReading = GetData(idx).channelA;
                    olderReading = GetData(idx - 1).channelA;
                }
                else
                {
                    newerReading = GetData(idx).channelB;
                    olderReading = GetData(idx - 1).channelB;
                }


                if ((risingEdge && newerReading >= triggerLevel && olderReading < triggerLevel) ||
                    (!risingEdge && newerReading <= triggerLevel && olderReading > triggerLevel))
                {
                    return idx;        // our search is done
                }
            }

            return -1;   // return a -1 if we haven't found a point to trigger on
        }
    }
}
